# =====================================================================================================
# Basic tool to dump content information and/or samples out of a DTPK file, 
# as well splicing in compatible PCM WAV files or YADPCM chunks.
# This tool is strongly based upon the previous research and work put together by
# kingshriek, the Shenmue team (especially Sappharad and LemonHaze), and many others.
# All wisdom is from them, all mistakes are from me.
# Clarifications always appreciated.

# In the interest of clear nomenclature, I'm going with track->sample(s)

# Update history located near end of file

# TODO list:
# * bugbug/todos as noted in the code
# * add BLANK option?
# * add delete option?
# * add joined-sample option?
# * add add-sample option?
# * add resampling to help with size concerns?
# * add volume change?
# * add rate change for samples reused at different rates (such as mvc2 Tron Bonne)?

# MIT License

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# =====================================================================================================

# =====================================================================================================
import os
import sys
import wave
import math
from struct import pack
from struct import unpack
from array import array
from enum import Enum
# =====================================================================================================

# =====================================================================================================

# Config options since there's a lot of data
show_detailed_file_info = False
show_header_info = False
show_detailed_data_tracks = False
show_detailed_data_playback = False
show_detailed_data_effects = False
show_detailed_data_samples = False
show_detailed_data_samples_full = False
dump_adpcm_samples = False
dump_pcm_samples_to_wav = False
dump_adpcm_samples_to_wav = False
reassembly_test = False
# FWIW, 6 * 156000 = 936000.  The MvC2 common match effects file (SE_COMN.BIN) is 931284.
# They worked hard to fit within the 2MB audio limit for Dreamcast, and if you're 
# injecting new samples you would want to adhere to that same minimalist approach.
mvc2_DTPK_size_limit = 156000

INVALID_SAMPLE_ID = 0xFFFF

# =====================================================================================================
# I'm unclear where the bcolors implementation sources from: it's referenced around the internet.
# Usage is as shown: make sure to close the presentation shift with ENDC.

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def PrintAsWarning(msg):
	print (bcolors.WARNING + msg + bcolors.ENDC)

def PrintAsFailure(msg):
	print (bcolors.FAIL + msg + bcolors.ENDC)

# =====================================================================================================
def DelayValuesTo44100SampleCount(unMaskedByte1, unMaskedByte2):
	# For a discussion of what these magic values are, see where we ingest the track composition data
	# Byte 1:
		# starts masked with 0x80
		# each odd +0x01 step is +0n5120 samples
		# each even +0x01 step is +0n6144 samples
		# a full 0x10 steps is +0n90112 samples (which is (8 * 5120) + (8 * 6144) conveniently enough)
	# Byte 2 is optional, and increments at chunks of 1024 samples every 0x17 steps
	delayInSamples = (int(math.floor(unMaskedByte1 / 2)) * 11264) + ((unMaskedByte1 % 2) * 5120)
	delayInSamples += int(math.floor(unMaskedByte2 / 23)) * 1024
	return delayInSamples

# =====================================================================================================
class DTPKTrackEntry:
	# A normal track entry looks like:
	# C0 DF 01 78 80 FF
	# In order:
	# Start_Track (C0) Sample_Type (DF) Sample_ID (01) Volume (78) End_Sample_Ref? (80) End_Track (FF)
	# Note: Maximum value of volume is 0x7f
	# For a Joined track such as Track 01 in SND_OPEN.BIN for cvs2:
	# C0 DF 01 78 00 8B 4E DF 02 78 80 FF
	# In order:
	# Start_Track (C0) Sample_Type (DF) Sample_ID (01) Volume (78) Join_With (00) Delay_Part1 (0b | 80) Opt_Delay_Part2 (4e)
	#    Sample_Type (DF) Sample_ID (02) Volume (78) End_Sample_Ref (80) End_Track (FF)
	# The Delay value is relative to 0:00, not the end of the previous sample.

	def __init__(self, entry_type, sample_id, volume, trailing_bytes):
		self.entry_type = entry_type
		self.sample_id = sample_id
		self.volume = volume
		self.trailing_bytes = trailing_bytes

	def Print(self):
		if (IsTrackEntryUsed(self.entry_type)):
			summary = '\t\t\tType: 0x%02x.  SPD: 0x%02x.  Volume: 0x%02x.' % (self.entry_type, self.sample_id, self.volume)

			extraBytes = ""

			# Track Entry errata:
			# Cammy-CVS2 (SND_PL0C) track 36 "joins" two non-existent samples (25, 26) using terminal byte(?) 80
			# not a join/append byte (00/01).  But it's an unused track, so the composition is irrelevant.
			if (self.trailing_bytes[0] & 0x80):
				extraBytes = ' Terminal byte: 0x%02x' % self.trailing_bytes[0]
			else:
				extraBytes = " Join bytes:"
				for val in self.trailing_bytes:
					extraBytes += " 0x%02x" % val

				if (self.trailing_bytes[0] == 0):
					extraBytes += '\n\t\t\t\t    Meaning: Join Next (0x%02x)' % self.trailing_bytes[0]
				else:
					extraBytes += '\n\t\t\t\t    Meaning: Replace With? (0x%02x)' % self.trailing_bytes[0]

				delayInSecs = 0
				if (len(self.trailing_bytes) > 1):
					unMaskedByte1 = self.trailing_bytes[1] ^ 0x80
					unMaskedByte2 = 0
					extraBytes += ' Delay Next: 0x%02x' % unMaskedByte1
					if (len(self.trailing_bytes) > 2):
						unMaskedByte2 = self.trailing_bytes[2]
						extraBytes += ' 0x%02x' % self.trailing_bytes[2]
					delayInSamples = DelayValuesTo44100SampleCount(unMaskedByte1, unMaskedByte2)
					extraBytes += ' (~%1.3fs)' % (delayInSamples / 44100) # I did all my math at 44100
				if (len(self.trailing_bytes) > 3):
					PrintAsFailure('ERROR: Unknown track entry data {}'.format(self.trailing_bytes))

			summary += extraBytes
		elif (self.entry_type == 0xa0):
			# Unused: our internal record of sample_id is INVALID_SAMPLE_ID here, and we stuff the unknown byte into Volume
			summary = '\t\t\tUNUSED: Type: 0x%02x.  Sample: N/A (0x%02x).  Terminal byte: 0x%02x' % (self.entry_type, self.volume, self.trailing_bytes[0])
		else:
			summary = '\t\t' + bcolors.FAIL + 'ERROR: Unknown type: 0x%02x.  Sample: N/A (0x%02x).  Terminal byte: 0x%02x' % (self.entry_type, self.sample_id, self.trailing_bytes[0]) + bcolors.ENDC
		print (summary)

# DTPKTrack entries: 5-8 bytes per track.
	# ~5 bytes per C0 A0 entry
	# ~6 bytes per C0 DF entry
	# First byte is usually a 0xc0 or 0xd? byte that doesn't appear to be relevant
	# A0 entries appear to indicate unused tracks
	# Penultimate byte appear to be 0x80 OR 0x81 (0x81 in sound effects files only so far)
	# 0xFF is always the terminal byte
	# This table can be used to join samples together!
	# For joins, based upon ~350 files, there is either two or three join bytes.
	# Consider AA BB [CC].
	# AA is wholly unknown but is either 00 (most often) or 01 (rarely)
	# BB appears to be 80 OR'd with a delay duration in 0.25s increments(?).  maybe in samples? unclear
	# CC is often present and seems to increase the delay duration
	# OK, spent some time generating variant DTPKs with different join delay values, and the values appear to be:
	# BB ^ 80 = variable length step.
		# each odd step is +0n5120 samples
		# each even step is +0n6144 samples
		# a full 0x10 steps is +0n90112 samples.
	# CC = floor(CC/0n23) * 0n1024 samples, so about 0.023s per increment
	# That's all speculative based upon an analysis of about 90 variant delays from 8000 (play at same time) to b96e, done with
	# both mvc2-dc and cvs2-dc files.
	# Somewhere between 0xb96e to 0xc96e we ... overflow(?) and the second/joined clip just doesn't play.

class DTPKTrack:
	def __init__(self, track_id, playback_id):
		self.track_id = track_id
		self.playback_id = playback_id
		self.entryList = []
		self.starting_offset = 0
		self.raw_bytes = []

	def AddRawEntryData(self, start_location, byte_array):
		read_offset = 1
		nTotalSPDsUsed = 0
		entry_type = (byte_array[read_offset:read_offset + 1])[0]
		self.raw_bytes = byte_array

		if (IsTrackEntryUsed(entry_type)):
			while (IsTrackEntryUsed(entry_type)):
				trailing_bytes = []
				read_offset += 1
				# sample to use
				sample_to_use = (byte_array[read_offset:read_offset+1])[0]
				read_offset += 1

				# Note that there can be a conflict(?) between the Track Composition Data sample value and 
				# the Track Playback Data sample value.  For at least flycast, the non-zero value of the two is ignored.
				# In the instance where both are non-zero, the Track Composition Data value is used.
				volume_maybe = (byte_array[read_offset:read_offset+1])[0]
				read_offset += 1
						
				while (((byte_array[read_offset:read_offset+1])[0] != 0xa0) and
						(not (IsTrackEntryUsed((byte_array[read_offset:read_offset+1])[0]))) and 
						((byte_array[read_offset:read_offset+1])[0] != 0xff)):
					# optional bytes
					cur_byte = (byte_array[read_offset:read_offset+1])[0]
					# first byte in the segment appears to be volume
					trailing_bytes.append(cur_byte)
					read_offset += 1
					# closure or join byte is 0xff
				#trailing_bytes = "%s." % (trailing_bytes)
				self.AddSampleUsage(start_location, entry_type, sample_to_use, volume_maybe, trailing_bytes)
				entry_type = (byte_array[read_offset:read_offset+1])[0]
		elif (entry_type == 0xa0):
			read_offset += 1
			# sample to use: unused for type 0xa0...?
			# either that or we don't include volume.  unclear what the magic byte here is for
			volume_maybe = (byte_array[read_offset:read_offset+1])[0]
			read_offset += 1
			trailing_bytes = []
			trailing_bytes.append((byte_array[read_offset:read_offset+1])[0])
			self.AddSampleUsage(start_location, entry_type, INVALID_SAMPLE_ID, volume_maybe, trailing_bytes)
		else:
			PrintAsFailure("ERROR! Track composition parsing error: Unknown entry type %02x at offset 0x%x from 0x%x" % (entry_type, read_offset, start_location))
		
		return nTotalSPDsUsed

	def AddSampleUsage(self, starting_offset, entry_type, sample_used, volume, trailing_bytes):
		self.starting_offset = starting_offset
		self.entryList.append(DTPKTrackEntry(entry_type, sample_used, volume, trailing_bytes))

	def GetSampleCount(self):
		sampleCount = 0
		for entry in self.entryList:
			if (entry.sample_id != INVALID_SAMPLE_ID):
				sampleCount += 1
		return sampleCount

	def Print(self):
		if (show_detailed_data_tracks):
			if (len(self.entryList)):
				if (IsTrackEntryUsed(self.entryList[0].entry_type)):
					# At least one
					print ('\t\tTrack %02u composition starting at 0x%x, using %u sequences:' % (self.track_id, self.starting_offset, len(self.entryList)))
				else:
					# Unused
					print ('\t\tTrack %02u composition starting at 0x%x (Unused):' % (self.track_id, self.starting_offset))
				for entry in self.entryList:
					entry.Print()
			else:
				PrintAsFailure('ERROR: Track %02u does not exist and should: this is likely a parsing error.' % (self.track_id))

class DTPKSequencerEntry:
	def __init__(self, sequencerChunkStart, chunkDescriptor, nTotalTracksSoFar, data_chunk):
		self.chunkDescriptor = chunkDescriptor
		self.countGroupItems = unpack('<I', data_chunk[0x00:0x04])[0] + 1
		self.sequencerChunkStart = sequencerChunkStart
		read_offset = self.chunkDescriptor & 0xFFFF

		groupEntryOffsets = []
		self.groupItems = []
		for iGroupEntry in range(self.countGroupItems):
			groupEntryOffsets.append(unpack('<I', data_chunk[4 + (4 * iGroupEntry):8 + (4 * iGroupEntry)])[0])

		for iGroupEntry in range(self.countGroupItems):
			chunkShiftedStart = groupEntryOffsets[iGroupEntry]
			if ((iGroupEntry + 1) < len(groupEntryOffsets)):
				chunkEnd = groupEntryOffsets[iGroupEntry + 1] - read_offset
			else:
				chunkEnd = len(data_chunk)

			current_track = DTPKTrack(iGroupEntry, nTotalTracksSoFar) 
			nTotalTracksSoFar += 1
			nCurrentGroupOffset = self.sequencerChunkStart + chunkShiftedStart
			current_track.AddRawEntryData(nCurrentGroupOffset, data_chunk[chunkShiftedStart - read_offset:chunkEnd])
			#nCurrentGroupOffset += (chunkShiftedStart - read_offset) - chunkEnd

			self.groupItems.append(current_track)

	def GetUnmaskedDescriptor(self):
		return self.chunkDescriptor >> 16

	def GetUnmaskedChunkOffset(self):
		return self.chunkDescriptor & 0xFFFF

	def GetTrackCount(self):
		return len(self.groupItems)

	def GetChunkLength(self):
		nChunkLength = 4 * (4 * len(self.groupItems))
		for groupEntry in self.groupItems:
			nChunkLength += len(groupEntry)
		return nChunkLength

	def PrintTableOffsets(self, this_chunk_offset):
		totalHandledEntryLength = 0
		for iPos, iCurEntry in enumerate(self.groupItems):
			offsetTableLoc = this_chunk_offset + (4 * iPos)
			entryLoc = this_chunk_offset + (4 * len(self.groupItems)) + totalHandledEntryLength
			entryRelativeLoc = entryLoc - self.sequencerChunkStart
			totalHandledEntryLength += len(iCurEntry.raw_bytes)
			print ('\t\tTrack data %2u (0x%x): Definition starts at 0x%x (0x%x from sequencer table start)' % (iPos, offsetTableLoc, entryLoc, entryRelativeLoc))

	def PrintTableEntries(self, nCurrentGroupOffset, nTotalTracksSoFar):
		for iGroupEntry in range(self.countGroupItems):
			self.groupItems[iGroupEntry].Print()

# For a DTPKSequencerChunk, consider the 44 track sample (PL0A_VOI 'rogue' from mvc2-dc):

# We have N sets of tables based upon how many sequencer groups there are.
# First at 0x1a8 is a 0-based 4 byte value indicating sequencer group count.  
# After that we have a 4 byte value per sequencer group.
# 0x0000ffff is the offset for that group within the sequencer table
# 0xffff0000 is the group_type OR'd with another value

# Each Group starts with a 4 byte value indicating the number of Tracks in that group.
# Each used Track is associated with a mostly unique sample playback definition in that table.
# VTennis does reuse SPDs in two instances (PACK1.bin and PACK8.bin) but in both cases
# the second usage is just a muted (vol 0x00) version of the SPD.
# Note that each SPD block is padded to a terminal 4 byte boundary

# So looking at PL0A / Rogue:
	# Sequencer 1 data starts after track count at 0x1b4:
		# Track Composition Read Offset data group 1 from 0x1b4-0x268
			# 4 bytes per track.
			# Each entry is the start_byte starting from start_sequencer_chunk for that specific entry
			# Consider:
			# c0 00 00 00 c6 00 00 00 cc 00 00 00
			# Track 0 starts at offset 0xc0 within the TCD.  Track 1 starts at offset 0x0c6.  Track 2 starts at offset 0xcc.
		# Track Composition Data from 0x268-0x358 // len 0xf0 / 240 .  Variable length per track, each section ends with FF
	# We are now in data for sequencer group 2 as of 0x358:
		# Track Composition Read Offset data group 1 0x358-0x40c // len b4 / 180
		# Track Composition Data from 0x40c-0x4fc // 0xf0 // 240

class DTPKSequencerChunk:
	def __init__(self, sequencerChunkStart, data_chunk):
		#self.data_chunk = data_chunk
		self.sequencerChunkStart = sequencerChunkStart
		countSequencerGroups = unpack('<I', data_chunk[0x00:0x04])[0] + 1
		self.SequencerGroups = []

		groupChunkOffsets = []
		for iGroup in range(countSequencerGroups):
			groupChunkOffsets.append(unpack('<I', data_chunk[4 + (4 * iGroup):8 + (4 * iGroup)])[0])

		nTotalTracksSoFar = 0
		for iGroup in range(countSequencerGroups):
			chunkMaskedStart = groupChunkOffsets[iGroup]
			if ((iGroup + 1) < len(groupChunkOffsets)):
				chunkMaskedEnd = groupChunkOffsets[iGroup + 1]
			else:
				chunkMaskedEnd = len(data_chunk)

			self.SequencerGroups.append(DTPKSequencerEntry(sequencerChunkStart, chunkMaskedStart, nTotalTracksSoFar, data_chunk[(chunkMaskedStart & 0xFFFF):(chunkMaskedEnd & 0xFFFF)]))
			nTotalTracksSoFar += self.SequencerGroups[iGroup].GetTrackCount()

	def GetGroupCount(self):
		return len(self.SequencerGroups)

	def GetTrackCount(self):
		nTotalTracksSoFar = 0
		for curGroup in self.SequencerGroups:
			nTotalTracksSoFar += curGroup.GetTrackCount()
		return nTotalTracksSoFar

	def FindSequencerEntry(self, playback_id):
		totalPlaybacksReferenced = 0

		group_id = track_id = 0xFFFF

		for iGroupPos, curGroup in enumerate(self.SequencerGroups):
			for iPos, curTrack in enumerate(curGroup.groupItems):
				totalPlaybacksReferenced += curTrack.GetSampleCount()
				if (playback_id < totalPlaybacksReferenced):
					track_id = iPos
					group_id = iGroupPos
					break
			if (track_id != 0xFFFF):
				break

		return group_id, track_id

	def GetBytes(self):
		outBytes = []
		nCountGroups = len(self.SequencerGroups)
		if (nCountGroups > 0):
			nCountGroups -= 1
			outBytes = array('B', nCountGroups.to_bytes(4, byteorder='little'))
			nOffsetToEntry = 4
			for iGroupPos, curGroup in enumerate(self.SequencerGroups):
				outBytes += array('B', curGroup.chunkDescriptor.to_bytes(4, byteorder='little'))
				nOffsetToEntry += 4

			for iGroupPos, curGroup in enumerate(self.SequencerGroups):
				nCountGroupItems = len(curGroup.groupItems) - 1
				outBytes += array('B', nCountGroupItems.to_bytes(4, byteorder='little'))
				# Walk past the group's track count
				nOffsetToEntry += 4
				# Walk past the group's track offset chunk
				nOffsetToEntry += (len(curGroup.groupItems) * 4)
				for iPos, curTrack in enumerate(curGroup.groupItems):
					outBytes += array('B', nOffsetToEntry.to_bytes(4, byteorder='little'))
					nOffsetToEntry += len(curTrack.raw_bytes)
				for iPos, curTrack in enumerate(curGroup.groupItems):
					outBytes += curTrack.raw_bytes

			# BUGBUG TODO: We're only accidentally handling padding here.  It works for right now
			# but we should be explicitly managing this.
		else:
			PrintAsFailure('ERROR: No sequencer groups present.')

		return outBytes

	def Print(self):
		print ('\n  ==================SEQUENCER DATA===================')
		print ('Group count:\t\t%u\t\t\t0x%x' % (self.GetGroupCount(), self.sequencerChunkStart))

		for iGroup, curGroup in enumerate(self.SequencerGroups):
			currentSequencerCommandLoc = self.sequencerChunkStart + 4 + (4 * iGroup)

			cmd_type = curGroup.GetUnmaskedDescriptor()
			cmd_offset = curGroup.GetUnmaskedChunkOffset()

			print ('Group %u description:\t0x%x\t\t\t0x%x' % (iGroup, cmd_type, currentSequencerCommandLoc))
			print ('\tData starts at: 0x%03x (0x%03x + 0x%03x)' % (cmd_offset + self.sequencerChunkStart, self.sequencerChunkStart, cmd_offset))
			if (cmd_type >> 8) == 0xA8:
				print ('\tType is:\tA8: Music')
			elif (cmd_type >> 8) == 0xA9:
				print ('\tType is:\tA9: Voice data')
			elif (cmd_type >> 8) == 0xAA:
				print ('\tType is:\tAA: Sound effects/other')
			else:
				# DOA2 uses AB for example
				print ('\tType is:\t(Unknown group type)')

		print ('\nSequencer group data (0x%03x):' % (self.SequencerGroups[0].GetUnmaskedChunkOffset() + self.sequencerChunkStart))

		nTotalTracksSoFar = 0
		for iGroup, curGroup in enumerate(self.SequencerGroups):
			startEntryOffsetsTable = self.sequencerChunkStart + curGroup.GetUnmaskedChunkOffset()
			startEntriesTable = startEntryOffsetsTable + 4
			print ('\tGroup %u count tracks:\t\t%u (0x%x)\t\t\t\t0x%x' % (iGroup, curGroup.GetTrackCount(), curGroup.GetTrackCount(), startEntryOffsetsTable))
			print ('\tGroup %u track composition offset data (starts at 0x%x)' % (iGroup, startEntriesTable))

			curGroup.PrintTableOffsets(startEntriesTable)

			nCurrentGroupOffset = startEntryOffsetsTable +  4 + (4 * curGroup.GetTrackCount())
			print ('\tGroup %u track composition data (starts at 0x%x):' % (iGroup, nCurrentGroupOffset))
			curGroup.PrintTableEntries(nCurrentGroupOffset, nTotalTracksSoFar)
			nTotalTracksSoFar += curGroup.GetTrackCount()

class DTPKPlaybackDataEntry:
	# Speculative format of the 64 byte track composition block 
	# Byte 00: track ID
	# Byte 01: Unknown, usually 80
	# Byte 02: sample bank ID used 
	# Byte 03: Unknown, only values 1 or 2 seem to be used
	# Byte 04: Unused?
	# Byte 05: ... offset?
	# Byte 06: Panning? Value 0x1f == left channel only.  Value 0x0f == right channel only.
		# Note that panning likely has 13 total steps
	# Byte 07: Volume?  Values 0x00-0x7f.  0x80 or higher uses default volume.
	# Byte 08: Volume?  Values 0x00-0xff.  Unsure of interaction with Byte 07
	# Byte 09: Rate/frequency?.  Often e8, f4
	# Byte 10: Unknown, usually 00.  Has no effect?
	# Byte 11: Unknown, usually 1f.  Has effect
	# Byte 12: Unknown, usually 00.  Has no effect?
	# Byte 13: Unknown, usually 1f.  Has no effect?
	# Byte 14: Unknown, usually 00.  Has no effect?
	# Byte 15: Unknown, usually 00.  Has effect
	# Byte 16: Unknown, usually 00.  Has no effect?
	# Byte 17: fd
	# Byte 18: 07
	# Byte 19: fd
	# Byte 20: 07
	# Byte 21: fd
	# Byte 22: 07
	# Byte 23: fd
	# Byte 24: 27
	# Byte 25: 00
	# Byte 26: 00
	# Byte 27: 00
	# Byte 28: 00
	# Byte 29: 00
	# Byte 30: 00
	# Byte 31: 7f has effect
	# Byte 32: 00
	# Byte 33: 00
	# Byte 34: 00
	# Byte 35: 00
	# Byte 36: 00
	# Byte 37: 00
	# Byte 38: 00
	# Byte 39: 00
	# Byte 40: 00
	# Byte 41: 00
	# Byte 42: 00
	# Byte 43: 00
	# Byte 44: 00
	# Byte 45: 00
	# Byte 46: 00
	# Byte 47: 80
	# Byte 48: 00
	# Byte 49: 00
	# Byte 50: 00
	# Byte 51: 00
	# Byte 52: 00
	# Byte 53: 00
	# Byte 54: 00
	# Byte 55: 00
	# Byte 56: 00
	# Byte 57: 00
	# Byte 58: 00
	# Byte 59: 00
	# Byte 60: 00
	# Byte 61: 00
	# Byte 62: 00
	# Byte 63: 00
	def __init__(self, file_offset, data_chunk):
		self.file_offset = file_offset
		self.data_chunk = data_chunk

	def GetPlaybackID(self):
		return (self.data_chunk[0:1])[0]

	def GetFileAbsolutePlaybackLocation(self):
		return self.file_offset

	def GetSampleUsed(self):
		# Note that there can be a conflict(?) between the Track Composition Data sample value and 
		# the Track Playback Data sample value.  For at least reDream, the non-zero value of the two is ignored.
		# In the instance where both are non-zero, the Track Composition Data value is used.
		return (self.data_chunk[2:3])[0]

	def GetRate(self):
		return (self.data_chunk[10:11])[0]

	def Print(self):
		print ('\t\tPlayback 0x%02x: Rate: 0x%02x (%5uHz). Sample %02u (0x%02x).\t0x%03x' % (self.GetPlaybackID(), self.GetRate(), TranslateDTPKRate(self.GetRate()), self.GetSampleUsed(), self.GetSampleUsed(), self.GetFileAbsolutePlaybackLocation()))

class DTPKPlaybackDataChunk:
	# The playback data chunk starts with a 0n80 / 0x50 sample playback definition header.
	# The only important here is at byte 0x10, which is the number of sample playback definitions.
	# We then have 0x40 bytes per playback definition

	# So in PL0A / Rogue we have Track Playback Data from 0x54c-0xf4c in 64byte chunks for those 44 tracks
	# The section is length 0xa00  / 0n2560
	def __init__(self, file_offset, data_chunk):
		self.file_offset = file_offset
		self.headerChunk = data_chunk[0:0x50]

		countPlaybackData = unpack('<H', data_chunk[0x10:0x12])[0] + 1

		self.PlaybackEntries = []
		data_offset = 0x50
		for iPlayback in range(countPlaybackData):
			self.PlaybackEntries.append(DTPKPlaybackDataEntry(self.file_offset + data_offset, data_chunk[data_offset:data_offset + 0x40]))
			data_offset += 0x40

	def GetBytes(self):
		outBytes = self.headerChunk
		for curPlayback in self.PlaybackEntries:
			outBytes += curPlayback.data_chunk
		return outBytes

	def Print(self):
		print ('\n  ================PLAYBACK DATA======================')
		print ('Sample playback data (64 bytes per sample) starting at 0x%x:' % self.file_offset)
		print ('\tSample playback entry count:\t%u (0x%02x)\t\t\t0x%x' % (len(self.PlaybackEntries), len(self.PlaybackEntries), self.file_offset + 0x10))

		for curPlaybackEntry in self.PlaybackEntries:
			curPlaybackEntry.Print()

class DTPKSampleDataEntry:
	# This parsing logic is based squarely upon https://github.com/Shenmue-Mods/DTPKUtil/blob/master/AM2-DTPK.txt
	# Any mistakes here are mine, any wisdom here is theirs
	def __init__(self, startOffsetChunk, sample_number,  header_chunk, full_sample_chunk):
		self.sample_number = sample_number
		self.def_offset = startOffsetChunk + 4 + (0x10 * sample_number)
		self.header_chunk = header_chunk
		self.sample_flags = unpack('<I', self.header_chunk[0:4])[0]
		sampleOffset = self.sample_flags & 0x007fffff
		
		sample_length = unpack('<I', self.header_chunk[12:16])[0]

		self.group_id = 0xFFFF
		self.track_id = 0xFFFF
		self.playback_id = 0xFFFF
		self.real_rate = 0

		self.sample_data = full_sample_chunk[(sampleOffset - startOffsetChunk):(sampleOffset - startOffsetChunk) + sample_length]

	def IsSampleADPCM(self):
		return (self.sample_flags & 0x01000000)

	def GetSampleOffset(self):
		return self.sample_flags & 0x007fffff

	def GetSampleQuality(self):
		return self.sample_flags &  0x00800000

	def GetSampleFormat(self):
		return self.sample_flags &  0x01000000

	def GetSampleUnknown(self):
		return self.sample_flags &  0x02000000

	def GetSampleLoopStart(self):
		return unpack('H', self.header_chunk[4:6])[0]
	
	def GetSampleLoopEnd(self):
		return unpack('H', self.header_chunk[6:8])[0]

	def GetSampleRate(self):
		# This value has not been used in any of the samples I've seen
		return unpack('<I', self.header_chunk[8:12])[0]

	def GetSampleLength(self):
		return unpack('<I', self.header_chunk[12:16])[0]

	def AssignUsageValues(self, group_id, track_id, actual_rate, playback_id):
		# These are grabbed from other chunks and are not present in the sample data
		self.group_id = group_id
		self.track_id = track_id
		self.playback_id = playback_id
		self.real_rate = actual_rate

	def GetFlagValueForNewLocation(self, location):
			if ((self.GetSampleFormat() == 0x1000000) and (self.GetSampleQuality() == 0)):
				return location | 0x1000000 | 0x2000000
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0)):
				return location
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0x800000)):
				return location | 0x0800000

	def Print(self):
		if (show_detailed_data_samples_full):
			print('\tSample %2u (0x%02x):\t\t\t\t\t\t0x%x' % (self.sample_number, self.sample_number, self.def_offset))
			print('\t\tLocation:\t0x%05x' % self.GetSampleOffset())

			if ((self.GetSampleFormat() == 0x1000000) and (self.GetSampleQuality() == 0)):
				print ('\t\tsample_format:  ADPCM  (0x1000000)')
				print ('\t\tsample_format:  4bit   (0x0000000)')
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0)):
				print ('\t\tsample_format:  PCM    (0x0000000)')
				print ('\t\tsample_quality: 16bit  (0x0000000)')
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0x800000)):
				print ('\t\tsample_format:  PCM    (0x0000000)')
				print ('\t\tsample_quality: 8bit   (0x0800000)')
			else:
				PrintAsFailure('ERROR!!! This is not a valid audio format for the AICA')
				print ('\t\tsample_format:  Unknown (0x%07x)', self.GetSampleFormat())
				print ('\t\tsample_quality: Unknown (0x%07x)', self.GetSampleQuality())

			print ('\t\tsample_unk:     0x%07x' % self.GetSampleUnknown())

			print ('\t\tloop_start:     0x%04x count samples' % self.GetSampleLoopStart())
			print ('\t\tloop_end:       0x%04x count samples' % self.GetSampleLoopEnd())
			print ('\t\tsample_rate:    0x%x' % self.GetSampleRate())
			print ('\t\tsample_length:  0x%08x bytes' % self.GetSampleLength())
			print ('\t\t\t(sample data is at 0x%x to 0x%x)' % (self.GetSampleOffset(), self.GetSampleOffset() + self.GetSampleLength()))
		else:
			print ('\tSample %02u (def at 0x%x)' % (self.sample_number, self.def_offset))

# =====================================================================================================
# This ADPCM->PCM region is thanks to Kode54, BERO, Megan Potter, and Sappharad.
# Sappharad's work is under the MIT License as included above.
# BERO is under GPL.
# Kode54 is under MAME GPL v2 or similar
# I'm unclear as to the licensing for KallistiOS and/or Megan Potter's work.
# This section is entirely a straight Python-ization of Sappharad's C# adpcm2wav code
# There is no new value or knowledge added here, it's just going to produce byte-identical
# output to that C# version.

#region AICA ADPCM decoding
	def adpcm2pcm(self, file_data, src, length):
		diff_lookup = array('i', [ 1, 3, 5, 7, 9, 11, 13, 15, -1, -3, -5, -7, -9, -11, -13, -15 ])
		index_scale = array('i', [ 0x0e6, 0x0e6, 0x0e6, 0x0e6, 0x133, 0x199, 0x200, 0x266 ])

		dst = bytearray(length * 4)
		dstLoc = 0
		cur_quant = 0x7f
		cur_sample = 0
		highNybble = False

		while (dstLoc < len(dst)):
			if (highNybble):
				shift1 = 4
			else:
				shift1 = 0
			delta = (file_data[src] >> shift1) & 0xf

			# ~2014 mame version by Kode54 / Metallic.  In testing Sapphard's appears to be more accurate for our needs.
			# Note that if you use this, get rid of the 254/256 line below
			#x = (cur_quant * diff_lookup[delta & 7]) / 8
			#x = min(x, 0x7fff)
			#if (delta & 8):
			#	x = -x

			x = cur_quant * diff_lookup[delta & 15]
			x = int(cur_sample + ((x + (self.signedToUnsigned(x, 4) >> 29)) >> 3))

			# standard cleanup
			cur_sample = min(max(x, -32768), 32767)
			cur_quant = (cur_quant * index_scale[delta & 7]) >> 8
			cur_quant = min(max(cur_quant, 0x7F), 0x6000)

			# standard byte packing
			dst[dstLoc] = (int(cur_sample) & 0xFF)
			dstLoc += 1
			dst[dstLoc] = ((int(cur_sample) >> 8) & 0xFF)
			dstLoc += 1

			cur_sample = int(cur_sample * 254 / 256)

			highNybble = not highNybble
			if (not highNybble):
				src += 1

		return dst
	
#endregion

	def ExtractIfYADPCM(self, filename_base):
		if ((self.GetSampleFormat() == 0x1000000) and (self.GetSampleQuality() == 0)):
			filename_out = self.GetFilename(filename_base, "yadpcm")
			sample_data = self.sample_data[0:self.GetSampleLength()]

			try:
				open(filename_out, 'wb').write(sample_data)
				print ('ADPCM export: %s' % filename_out)
				return 1
			except:
				PrintAsFailure('ERROR: filename \'%s\' cannot be written to currently.' % filename_out)

			return 1
		return 0

	def ExtractPCMToWAV(self, filename_base):
		# Note that this sketches in nonPCM export, but that won't work because we don't have a VFW decoder
		# The Shenmue team at https://github.com/Shenmue-Mods/DTPKUtil/blob/master/DTPKutil/DtpkFile.cs
		# have a fairly functional adpcmTopcm converter, but it does skewed conversion for 
		# our files
		# VGMSteam's code at https://github.com/vgmstream/vgmstream/blob/master/src/coding/yamaha_decoder.c
		# is also interesting
		if (self.GetSampleFormat() == 0x0000000) or (dump_adpcm_samples_to_wav):
			filename_out = self.GetFilename(filename_base, "wav")
		
			riffHeader = b'RIFF'
			riffType = b'WAVE'
			chunkFmtName = b'fmt '

			# 2b formattag
			# So uh MSFT's formattag registration tables are a little screwy.  In theory
			# the correct formattag might be 0n20 / 0x14 for Yamaha ADPCM, but I haven't
			# seen a VFW decoder that works so who actually knows.  But hey: use 0n20 if 
			# want to be exporting ADPCM WAV
			# PCM == 0x01
			wFormatTag = b'\x01\x00'

			# 2b number of channels
			# ... everything shall be mono here
			nChannels = b'\x01\x00'

			# 4b sample rate: use the DTPK rate value
			# 4b bytes per second: likewise
			rate_to_use = self.GetSampleRate()
			if (rate_to_use == 0):
				rate_to_use = self.real_rate

			frequency = TranslateDTPKRate(rate_to_use)
			nSamplesPerSec = nAvgBytesPerSec = frequency.to_bytes(4, byteorder='little')

			# 2b block align
			nBlockAlign = b'\x02\x00'
			# 2b bits per sample

			fIsHalfQuality = False
			if ((self.GetSampleFormat() == 0x1000000) and (self.GetSampleQuality() == 0)):
				# We convert ADPCM 4bit to PCM 16bit for export
				nBitsPerSample = b'\x10\x00'
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0)):
				# PCM 16bit
				nBitsPerSample = b'\x10\x00'
			elif ((self.GetSampleFormat() == 0) and (self.GetSampleQuality() == 0x800000)):
				# PCM 8bit
				nBitsPerSample = b'\x08\x00'
				fIsHalfQuality = True

			exData = b''

			# You only need this if you're exporting out ADPCM-as-ADPCM within WAV.  But again: no VFW decoder
			# for that that I'm aware of
			#if (self.format == 0x1000000): #ADPCM
			#	exData = b'\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' 

			chunkFmtData = wFormatTag + nChannels + nSamplesPerSec + nAvgBytesPerSec + nBlockAlign + nBitsPerSample + exData

			chunkFmtSize = pack('l', len(chunkFmtData))

			chunkDataName = b'data'

			fWasYADPCM = False
			if (self.GetSampleFormat() == 0x1000000):
				fWasYADPCM = True
				chunkDataData = self.adpcm2pcm(self.sample_data, 0, self.GetSampleLength())
			else:
				chunkDataData = self.sample_data[0:self.GetSampleLength()]

			if fIsHalfQuality:
				# Xor the data here and I don't know why
				for iPos, iCurr in enumerate(chunkDataData):
					chunkDataData[iPos] = chunkDataData[iPos] ^ 0x80

			chunkDataSize = pack('l', len(chunkDataData))

			riffBody = riffType + chunkFmtName + chunkFmtSize + chunkFmtData + chunkDataName + chunkDataSize + chunkDataData

			fileSize = pack('l', len(riffBody))

			try:
				open(filename_out, 'wb').write(riffHeader + fileSize + riffBody)
				if (fWasYADPCM):
					print ('PCM export: %s\t(from ADPCM)' % filename_out)
				else:
					print ('PCM export: %s' % filename_out)
				return 1
			except:
				PrintAsFailure('ERROR: filename \'%s\' cannot be written to currently.' % filename_out)
		return 0

	def GetFilename(self, filename_base, extension):
		if (self.group_id == 0xffff):
			strGroupId = "NA"
		else:
			strGroupId = '%02u' % self.group_id

		if (self.track_id == 0xffff):
			strTrackId = "NA"
		else:
			strTrackId = '%02u' % self.track_id

		if (self.playback_id == 0xffff):
			strPlaybackId = "NA"
		else:
			strPlaybackId = '%02u' % self.playback_id

		#create file
		# bugbug todo: It may be much more useful to have the frequency in the filename rather than the internal rate value
		if (self.GetSampleRate() == 0):
			filename_out = ('%s_%s_%s_SPD_%s_Sample_%02X_Rate_%02X.%s' % (filename_base, strGroupId, strTrackId, strPlaybackId, self.sample_number, self.real_rate, extension))
		else:
			filename_out = ('%s_%s_%s_SPD_%s_Sample_%02X_Rate_%02X.%s' % (filename_base, strGroupId, strTrackId, strPlaybackId, self.sample_number, self.GetSampleRate(), extension))

		return filename_out

	# This needed utility code comes from hl037_ on stackoverflow
	def signedToUnsigned(self, n, byte_count): 
		return int.from_bytes(n.to_bytes(byte_count, 'little', signed=True), 'little', signed=False)

class DTPKSampleDataChunk:
	def __init__(self, file_offset, data_chunk):
		self.file_offset = file_offset
		self.countSamples = unpack('<I', data_chunk[0:4])[0] + 1

		self.SampleEntries = []

		data_offset = 4
		for iSample in range(self.countSamples):
			self.SampleEntries.append(DTPKSampleDataEntry(self.file_offset, iSample, data_chunk[data_offset:data_offset + 0x10], data_chunk))
			data_offset += 0x10

	def GetCountADPCMSamples(self):
		nTotalCount = 0
		for curEntry in self.SampleEntries:
			if (curEntry.IsSampleADPCM()):
				nTotalCount +=1
		return nTotalCount

	def GetCountPCMSamples(self):
		nTotalCount = 0
		for curEntry in self.SampleEntries:
			if (not curEntry.IsSampleADPCM()):
				nTotalCount +=1
		return nTotalCount

	def ExtractIfYADPCM(self, filename_base):
		nTotalCount = 0
		for curEntry in self.SampleEntries:
			nTotalCount += curEntry.ExtractIfYADPCM(filename_base)
		return nTotalCount

	def ExtractPCMToWAV(self, filename_base):
		nTotalCount = 0
		for curEntry in self.SampleEntries:
			nTotalCount += curEntry.ExtractPCMToWAV(filename_base)
		return nTotalCount

	def AssignFirstSequencerEntries(self, chunkSequencer, chunkPlaybackData):
		actual_rate = 0
		playback_id = 0xFFFF
		track_id = 0xFFFF
		group_id = 0xFFFF

		for curSampleEntry in self.SampleEntries:
			for curPlayback in chunkPlaybackData.PlaybackEntries:
				if (curPlayback.GetSampleUsed() == curSampleEntry.sample_number):
					actual_rate = curPlayback.GetRate()
					playback_id = curPlayback.GetPlaybackID()

					group_id, track_id = chunkSequencer.FindSequencerEntry(playback_id)

					break

			if (track_id == 0xFFFF):
				PrintAsWarning('Errata: playback ID for sample 0x%02x not found.  This file is using an unexpected structure.' % curSample)
			curSampleEntry.AssignUsageValues(group_id, track_id, actual_rate, playback_id)

	def GetBytes(self):
		nCountSamples = len(self.SampleEntries)
		nCountSamples -= 1
		outBytes = array('B', nCountSamples.to_bytes(4, byteorder='little'))

		for curSampleEntry in self.SampleEntries:
			outBytes += curSampleEntry.header_chunk

		for curSampleEntry in self.SampleEntries:
			outBytes += curSampleEntry.sample_data

		# Handle padding
		remnant = len(outBytes) % 4
		if (remnant != 0):
			outBytes += array('B', bytes(remnant))

		return outBytes

	def Print(self):
		print ('\n  ==============SAMPLE METADATA=====================')
		print ('Sample metadata:')
		print ('\tSample count:\t\t%2u (0x%02x)\t\t\t\t0x%x' % (self.countSamples, self.countSamples, self.file_offset))

		for curEntry in self.SampleEntries:
			curEntry.Print()

class DTPKEffectsChunk:
	# Thanks to Vincent for figuring this out!
	# At the start of the Effects chunk is the 4 byte header.  That is the number of effect groups.
	# Each effect group chunk has a 0x24 byte header describing PAN and VOLUME values for each channel.
		# PAN values:
		# 0x00 = Center
		# 0x01-0xF = Right Pan 1-15
		# 0x11-0x1F = Left Pan 1-15

		# VOLUME values:
		# 0x00-0x0F = Volume 0-15
	# This is followed by the .FPD chunk which is 0xc00 bytes.  That chunk handles reverb, fades, chorus, pitch, etc.

	def __init__(self, file_offset, data_chunk):
		self.file_offset = file_offset
		self.data_chunk = data_chunk

	def GetBytes(self):
		return self.data_chunk

	def GetStringForPanValue(self, value):
		if (value == 0x00):
			return "Center"
		elif (value & 0x10):
			return "Left %2u" % (value & 0x0f)
		else:
			return "Right %2u" % value

	def GetEffectGroupsCount(self):
		return unpack('<I', self.data_chunk[0x00:0x04])[0] + 1

	def Print(self):
		if (self.data_chunk):
			print ('\n  ================EFFECTS DATA=======================')
			print ('Effects group count:\t\t%2u (0x%02x)\t\t\t\t0x%x' % (self.GetEffectGroupsCount(), self.GetEffectGroupsCount(), self.file_offset))

			nCurrentOffset = 4
			for iCurGroup in range(self.GetEffectGroupsCount()):
				print ('\tEffects group 0x%02x\t\t\t\t\t\t0x%x' % (iCurGroup, self.file_offset + nCurrentOffset))
				for iCurChannel in range(18):
					bytePan = (self.data_chunk[nCurrentOffset:nCurrentOffset + 1])[0]
					byteVolume = (self.data_chunk[nCurrentOffset + 1:nCurrentOffset + 2])[0]

					if ((bytePan > 0x1f) or (byteVolume > 0x0f)):
						PrintAsFailure('ERROR: Unknown effects value present at 0x%x' % (self.file_offset + nCurrentOffset))

					print ('\t\tChannel %2u: PAN %8s, Volume: %2u (0x%02x, 0x%02x)\t0x%x' % (iCurChannel, self.GetStringForPanValue(bytePan), byteVolume, bytePan, byteVolume, self.file_offset +nCurrentOffset))
					nCurrentOffset += 2
				nCurrentOffset += 0xc00

class DTPKUnhandledChunk:
	def __init__(self, data_chunk):
		self.data_chunk = data_chunk

	def GetBytes(self):
		return self.data_chunk

class DTPKHeaderChunk:
	def __init__(self, data_chunk):
		self.data_chunk = data_chunk

	def GetBytes(self):
		return self.data_chunk

	def GetDTPKId(self):
		return unpack('<I', self.data_chunk[0x04:0x08])[0]

	def GetFileSize(self):
		return unpack('<I', self.data_chunk[0x08:0x0c])[0]

	def SetFileSize(self, newValue):
		self.data_chunk[0x08:0x0c] = array('B', newValue.to_bytes(4, byteorder='little'))

	def WarnIfFileSizeValueIsWrong(self, filesize_actual):
	# Note we may adjust this size for write-out if we splice in samples
		if (self.GetFileSize() != filesize_actual):
			# This will cause unexpected errors, so mention it to the user.
			PrintAsFailure('ERROR! Filesize in file is listed as 0x%x, but actual filesize is 0x%x.  This file may not play back correctly.' % (self.GetFileSize(), filesize_actual))

	def GetUnknownChunkOffset(self):
		return unpack('<I', self.data_chunk[0x28:0x2c])[0]

	def GetSequencerChunkOffset(self):
		return unpack('<I', self.data_chunk[0x2c:0x30])[0]

	def GetProgramChunkOffset(self):
		return unpack('<I', self.data_chunk[0x30:0x34])[0]

	def GetOptionalEffectsChunkOffset(self):
		return unpack('<I', self.data_chunk[0x38:0x3c])[0]

	def GetSampleChunkOffset(self):
		return unpack('<I', self.data_chunk[0x3c:0x40])[0]

	def Print(self):
		print ('\n  ====================HEADER DATA====================')
		print ('Signature:\t\tDTPK\t\t\t0x00')
		print ('DTPK id:\t\t0x%x\t\t\t0x04' % self.GetDTPKId())
		print ('Filesize:\t\t0x%x (0n%u)\t0x08' % (self.GetFileSize(), self.GetFileSize()))
		print ('Header size:\t\t\t\t\t\tLength: 0x%x bytes' % (self.GetUnknownChunkOffset()))
		print ('Unknown chunk at:\t0x%x\t\t\t0x28\tLength: 0x%x bytes' % (self.GetUnknownChunkOffset(), (self.GetSequencerChunkOffset() - self.GetUnknownChunkOffset())))
		print ('Seq data at:\t\t0x%x\t\t\t0x2c\tLength: 0x%x bytes' % (self.GetSequencerChunkOffset(), (self.GetProgramChunkOffset() - self.GetSequencerChunkOffset())))
		nextChunkLoc = self.GetOptionalEffectsChunkOffset()
		if (self.GetOptionalEffectsChunkOffset() == 0):
			nextChunkLoc = self.GetSampleChunkOffset()
		print ('Playback data at:\t0x%x\t\t\t0x30\tLength: 0x%x bytes' % (self.GetProgramChunkOffset(), (nextChunkLoc - self.GetProgramChunkOffset())))
		unknownChunkLength = 0
		if (self.GetOptionalEffectsChunkOffset() != 0):
			unknownChunkLength = self.GetSampleChunkOffset() - self.GetOptionalEffectsChunkOffset()
		print ('Effects chunk at:\t0x%x\t\t\t0x38\tLength: 0x%x bytes' % (self.GetOptionalEffectsChunkOffset(), unknownChunkLength))
		print ('Sample data at:\t\t0x%x\t\t\t0x3c\tLength: 0x%x bytes' % (self.GetSampleChunkOffset(), (self.GetFileSize() - self.GetSampleChunkOffset())))

# =====================================================================================================
def IsTrackEntryUsed(entryType):
	if ((entryType == 0xdc) or # seen in cvs2
	    (entryType == 0xdd) or # seen in cvs2
	    (entryType == 0xde) or # seen in cvs2
		(entryType == 0xdf)):  # seen in mvc2
		return True
	return False

# =====================================================================================================
def TranslateDTPKRate(frequency):
	#DTPK rate values can be found in these examples:
	# Unknown values (bugbug todo)
	# 01 doa fight_se
		#  78/0x4e, sample 0x23 - track 17 join?
		# 120/0x78, sample 0x21 - track 62 join?
		# 123/0x7b, sample 0x16 - track 73 join?
		# 148/0x94, sample 0x29 - track 78 join?
	# 02 (doa2 fight_se km_c2/ks_c2/kt_end/tg_c123): ?  if this is correct, this is unknown.
		# 25/0x19, sample 0x28 - part of track 8
		# 74/0x4a sample 0x29
	# 04 cvs2 SND_PL2A (Hibiki): track 49/0x31, sample 39/0x27, also track 103/0x67
		# confirmed in file
		# seems to be about 44100-45000?
	# 05 doa fight_se
		# 75/0x4b, sample 0x29

	# Unused value...?  Not going to make this the new zero_value because if there is usage of 0xd1-0xdd I'd like to test them first
	# d1 DOA voice_le voice_lu: unused - empty samples

	# "Known" values:
	# 0x0 44100: doa2
	
	# 0xde 6000: mvc2 pl3a kobun/servbot SPD 0x1f sample 0x1a track 39 (slowdown of 38)
	# 0xdf 6500: unused? not seen in content yet
	# 0xe0 7000: unused? not seen in content yet
	# 0xe1 7500: mvc2 pl10_voi tron bonne, spd 0x17, 0x18, 0x41, 0x42 - sample 0x20 0x15, track 30 (slowdown of 29) and 32
	# 0xe2 8000: mvc2 guile 02
	# 0xe3 8500: speculative: needs verification
	# 0xe4 9000: speculative: seems good for mvc2 sakura, abyss
	# 0xe5 9500: speculative: needs verification
	# 0xe6 10000: cvs2 rolento
	# 0xe7 10500: speculative: seems good for mvc2 sakura
	# 0xe8 11025: most
	# 0xe9 12000: cvs2 rolento
	# 0xea 12500: speculative: needs verification
	# 0xeb 13000: speculative: seems good for mvc2 bonerine
	# 0xec 14000: speculative: needs verification
	# 0xed 15000: speculative: needs verification
	# 0xee 16000: speculative: seems good for mvc2 storm, jin
	# 0xef 17000: speculative: seems good for doa2 voice_ba
	# 0xf0 18000: speculative: seems good for mvc2 storm, spiral
	# 0xf1 19000: speculative: seems good for doa2 fight_se/voice_kl, cvs2 guile
	# 0xf2 20000: speculative: seems good for vtennis pack 8
	# 0xf3 21000: speculative: seems good for vtennis pack 8
	# 0xf4 22050: lots
	# 0xf5 23000: cvs2 dhalsim
	# 0xf6 24000: cvs2 vega / vtennis pack 4
	# 0xf7 26000: speculative: seems good for vtennis pack4, doa2 km_c1
	# 0xf8 28000: speculative: seems good for doa2 fight_se
	# 0xf9 30000: speculative: needs verification
	# 0xfa 32000: cvs2 cammy
	# 0xfb 34000: virt tennis pack0
	# 0xfc 35000: speculative: needs verification
	# 0xfd 38000: speculative: seems good for vtennis pack0
	# 0xfe 40000: speculative: seems good for doa2 fight_se
	# 0xff 42000: speculative: seems good for doa2 fight_se

	rateMap = [ 6000, 6500, 7000, 7500, 8000, # de...
				8500, 9000, 9500, 10000, 10500, 11025, # e3...
				12000, 12500, 13000, 14000, 15000, 16000, 17000, # e9...
				18000, 19000, 20000, 21000, 22050, 23000, 24000, # f0...
				26000, 28000, 30000, 32000, 34000, 35000, 38000, # f7...
				40000, 42000, # 0xfe 0xff
				# Note that everything above 44100 is highly speculative
				44100, 45000, 46000, 47000, 48000, 49000 # 0x00 0x01 0x02 0x03 0x04 0x05
				]

	# This is lowest_known_DTPK_rate_value
	zero_val = 0xde
	max_low_value = 0x05

	if (frequency < 0x100): # DTPK rate to frequency
		table_index = frequency
		if ((table_index > max_low_value) and (table_index < zero_val)):
			if (table_index == 0xd1):
				PrintAsWarning('Errata: unusual rate 0xd1, seems only to be used for empty samples.  Possibly 0hz?')
			else:
				PrintAsFailure('ERROR: unknown DTPK rate value 0x%02x: Please let Preppy know about this file so we can figure it out.' % frequency)
			return 22050
		if (table_index <= max_low_value):
			table_index += 0x100

		table_index -= zero_val

		return rateMap[table_index]
	else: # frequency to DTPK rate
		for iPos, iVal in enumerate(rateMap):
			if (iVal == frequency):
				return (iPos + zero_val) & 0xFF
		PrintAsWarning('Warning: frequency %u is not supported in DTPK.  For best results please resample this to a supported rate' % frequency)
		# No exact match: use a close-ish one
		for iPos, iVal in enumerate(rateMap):
			if (iPos == 0):
				continue
			if (frequency <= iVal):
				return iVal
		return 0
	
# =====================================================================================================
class InjectableSampleType(Enum):
	PCM16  = 1
	PCM8   = 2
	ADPCM4 = 3

# =====================================================================================================
class InjectableSample:
	samplePosition = 0
	sampleType = InjectableSampleType.PCM16
	rate = 22050
	dataObject = []

	def __init__(self, position, type, waveData, rate):
		self.samplePosition = position
		self.sampleType = type
		self.dataObject = waveData

		# If we get a DTPK rate value, convert it to the standard playback rate
		if (rate < 0x100):
			rate = TranslateDTPKRate(rate)

		self.rate = rate

	def GetHeader(self, location):
		# Rate seems to be unused in this data chunk
		sampleRate = 0
		sampleLength = len(self.dataObject)
		if (self.sampleType == InjectableSampleType.ADPCM4):
			flags = location | 0x1000000 | 0x02000000
			loopEnd = sampleLength * 2
			# 28 is a speculative magic number here as seen in cvs2 adpcm samples
			# and as such could be wrong
			loopStart = loopEnd - 28
		elif (self.sampleType == InjectableSampleType.PCM8):
			flags = location | 0x0800000
			loopStart = 0
			loopEnd = sampleLength
		elif (self.sampleType == InjectableSampleType.PCM16):
			flags = location
			loopStart = 0
			loopEnd = int(sampleLength / 2)

		outBytes = bytearray(16)
		outBytes[0:4] = array('B', flags.to_bytes(4, byteorder='little'))
		outBytes[4:6] = array('B', loopStart.to_bytes(2, byteorder='little'))
		outBytes[6:8] = array('B', loopEnd.to_bytes(2, byteorder='little'))
		outBytes[8:12] = array('B', b'\x00\x00\x00\x00')
		outBytes[12:16] = array('B', sampleLength.to_bytes(4, byteorder='little'))
		return outBytes

# =====================================================================================================
def ImportAndAddADPCMFile(adpcmFilename, sampleList, samplePosition, sampleRate):
	# This is just a raw data chunk
	dataChunk = array('B', open(adpcmFilename, 'rb').read())
	sampleType = InjectableSampleType.ADPCM4
	sampleList.append(InjectableSample(samplePosition, sampleType, dataChunk, sampleRate))
	return True

# =====================================================================================================
def ImportAndAddWAVFile(wavFilename, sampleList, samplePosition):
	try:
		waveObject = wave.open(wavFilename, 'rb')
	except:
		PrintAsFailure("Not a supported WAV: only raw PCM is natively supported in Python")
		return False

	isValid = True
	try:
		nChannel = waveObject.getnchannels()
		if (nChannel != 1):
			print ('Only mono WAV is supported')
			raise

		nSampleWidth = waveObject.getsampwidth()
		if (nSampleWidth == 1):
			sampleType = InjectableSampleType.PCM8
		elif (nSampleWidth == 2):
			sampleType = InjectableSampleType.PCM16
		else:
			PrintAsFailure('ERROR: sample width is {}: only 1 or 2 is supported'.format(nSampleWidth))
			raise

		nFrameRate = waveObject.getframerate()
		if (TranslateDTPKRate(nFrameRate) == 0xbadf00d):
			PrintAsFailure('ERROR: framerate is {}: only a few rates are currently supported, check code'.format(nFrameRate))
			raise

		nFrames = waveObject.getnframes()

		flDuration = nFrames / nFrameRate

		dataChunk = waveObject.readframes(nFrames)

		if (nSampleWidth == 1):
			# Unknown XOR as noted below
			fixedAudio = bytearray(len(dataChunk))
			for iPos, iCurr in enumerate(dataChunk):
				fixedAudio[iPos] = dataChunk[iPos] ^ 0x80
			dataChunk = fixedAudio

		print ('Requesting to inject file {}:\n\tChannels {} SampWidth {} Framerate {} Count frames {} Duration {:2.2}s\n'.format(wavFilename, nChannel, nSampleWidth, nFrameRate, nFrames, flDuration) )

		sampleList.append(InjectableSample(samplePosition, sampleType, dataChunk, nFrameRate))

	except:
		PrintAsFailure('ERROR: Unable to load this WAV file.')
		isValid = False

	waveObject.close()

	return isValid

# =====================================================================================================


# =====================================================================================================
# Display data in the DTPK file
def dtpkparse(dtpkfile, outFile, changedSamples):
	try:
		dtpkbytes = array('B', open(dtpkfile, 'rb').read())
	except:
		PrintAsFailure('ERROR: \'%s\' could not be opened.' % dtpkfile)
		sys.exit(0)

	if (dtpkbytes[:4].tobytes() != b'DTPK'):
		print ('Skipping %s - not a valid DTPK sound file' % dtpkfile )
		return

	fIsShowingAnyInfo = show_detailed_file_info or show_header_info or show_detailed_data_tracks or show_detailed_data_playback or show_detailed_data_samples or show_detailed_data_samples_full
	fIsProbablyMvC2CharacterFile =  (dtpkfile.upper().startswith("PL") and dtpkfile.upper().endswith("_VOI.BIN"))
	filesize_actual = len(dtpkbytes)
	(filename_base, filename_ext) = os.path.splitext(dtpkfile)

	# Header size is typically 0x124 for most files or 0x230 for VTennis.
	endHeader = unpack('<I', dtpkbytes[0x28:0x2c])[0]
	if (endHeader == 0):
		PrintAsFailure('ERROR: This file is missing the first post-header chunk.  We should look into this file.')
		endHeader = unpack('<I', dtpkbytes[0x2C:0x30])[0]

	# Header, usually about 0x124 or 0x230 bytes.
	chunkHeader = DTPKHeaderChunk(dtpkbytes[0:endHeader])
	# Unknown chunk, always 0x84 bytes.
	chunkUnknownData = DTPKUnhandledChunk(dtpkbytes[endHeader:chunkHeader.GetSequencerChunkOffset()])
	
	chunkHeader.WarnIfFileSizeValueIsWrong(filesize_actual)

	if (chunkHeader.GetSequencerChunkOffset() == 0):
		print ('Skipping %s - no sequencer data found' % dtpkfile)
		return

	chunkSequencer = DTPKSequencerChunk(chunkHeader.GetSequencerChunkOffset(), dtpkbytes[chunkHeader.GetSequencerChunkOffset():chunkHeader.GetProgramChunkOffset()])

	# Our chunk parsing has to account for the optional Unknown chunk, only present so far in VTennis files.
	startChunkAfterPlaybackChunk = chunkHeader.GetOptionalEffectsChunkOffset()
	if (startChunkAfterPlaybackChunk == 0):
		# Just set this to zero length
		startChunkAfterPlaybackChunk = chunkHeader.GetSampleChunkOffset()

	chunkPlaybackData = DTPKPlaybackDataChunk(chunkHeader.GetProgramChunkOffset(), dtpkbytes[chunkHeader.GetProgramChunkOffset():startChunkAfterPlaybackChunk])
	chunkEffectsData = DTPKEffectsChunk(chunkHeader.GetOptionalEffectsChunkOffset(), dtpkbytes[startChunkAfterPlaybackChunk:chunkHeader.GetSampleChunkOffset()])
	chunkSampleData = DTPKSampleDataChunk(chunkHeader.GetSampleChunkOffset(), dtpkbytes[chunkHeader.GetSampleChunkOffset():chunkHeader.GetFileSize()])

	# Note that we look up sample count a little early so we know how many sample metadata entries we will need to read
	count_samples = unpack('<I', dtpkbytes[chunkHeader.GetSampleChunkOffset():chunkHeader.GetSampleChunkOffset() + 4])[0] + 1

	countPCMSamples = chunkSampleData.GetCountPCMSamples()
	countADPCMSamples = chunkSampleData.GetCountADPCMSamples()

	if (show_header_info):
		print ('Analyzing %s:' % dtpkfile)
		chunkHeader.Print()
	if (show_detailed_data_tracks):
		chunkSequencer.Print()
	if show_detailed_data_playback:
		chunkPlaybackData.Print()
	if show_detailed_data_effects:
		chunkEffectsData.Print()
	if show_detailed_data_samples:
		chunkSampleData.Print()

	chunkSampleData.AssignFirstSequencerEntries(chunkSequencer, chunkPlaybackData)

	countADPCMExports = 0
	countWAVExports = 0
	if (dump_adpcm_samples):
		countADPCMExports = chunkSampleData.ExtractIfYADPCM(filename_base)
	if (dump_pcm_samples_to_wav):
		countWAVExports = chunkSampleData.ExtractPCMToWAV(filename_base)

	if (reassembly_test):
		#chunkSampleData.HandleInjectedSamples(foo)
		#chunkHeader.GetBytes()
		#chunkUnknownData.GetBytes()
		#chunkSequencer.GetBytes()
		#chunkPlaybackData.GetBytes()
		#chunkEffectsData.GetBytes()
		#chunkSampleData.GetBytes()
		reassemblyFileName = "Corrupt.BIN"
		testFile =   array('B', chunkHeader.GetBytes() + chunkUnknownData.GetBytes() + chunkSequencer.GetBytes() + chunkPlaybackData.GetBytes() + chunkEffectsData.GetBytes() + chunkSampleData.GetBytes())
		open(reassemblyFileName, "wb").write(testFile)
		print ('\nFile \'%s\' created.  0x%x bytes.' % (reassemblyFileName, len(testFile)))



	# REMAINING OLDNESS
	# OK, we now have all the sample definition data: do we want to update it?
	if (len(changedSamples) != 0):
		dtpkOutBytes = array('B', dtpkbytes)
		currentSampleStartingOffset = chunkHeader.GetSampleChunkOffset() + 4
		injectionOffset = currentSampleStartingOffset + (16 * count_samples)

		for curSampleEntry in chunkSampleData.SampleEntries:
			fMustUpdateHeader = (len(changedSamples) != 0)

			for curUpdatedSample in changedSamples:
				if (curUpdatedSample.samplePosition == curSampleEntry.sample_number):
					curSampleSize = len(curUpdatedSample.dataObject)
					print ('Changing sample %u to %s len 0x%x' % (curUpdatedSample.samplePosition, curUpdatedSample.sampleType, curSampleSize))

					#currentSampleStartingOffset = chunkHeader.GetSampleChunkOffset() + 4 + (curUpdatedSample.samplePosition * 0x10)
				#injectionOffset = currentSampleStartingOffset + (16 * count_samples)
					print ('\tRebuilding sample definition entry at 0x%x' % currentSampleStartingOffset)
					dtpkOutBytes[currentSampleStartingOffset:currentSampleStartingOffset + 16] = array('B', curUpdatedSample.GetHeader(injectionOffset))

					print ('\tSample injected inline at 0x%x to 0x%x, size change of %u' % (injectionOffset, injectionOffset + curSampleSize, curSampleSize - curSampleEntry.GetSampleLength()))
					dtpkOutBytes[injectionOffset:injectionOffset + curSampleSize] = array('B', curUpdatedSample.dataObject)
					injectionOffset += curSampleSize
					remnant = injectionOffset % 4
					if (remnant != 0):
						print ('\tAdding %u bytes padding (end of sample was 0x%x)' % (remnant, injectionOffset))
						dtpkOutBytes[injectionOffset:injectionOffset + remnant] = array('B', bytes(remnant))
						injectionOffset += remnant

					print ('\tUpdating sampling rate...')
					previousRate = INVALID_SAMPLE_ID
					warnedOnce = False
					for curPlayback in chunkPlaybackData.PlaybackEntries:
						if (curPlayback.GetSampleUsed() == curUpdatedSample.samplePosition):
							thisOldRate = dtpkOutBytes[curPlayback.GetFileAbsolutePlaybackLocation() + 10:curPlayback.GetFileAbsolutePlaybackLocation() + 11][0]
							nUpdatedRate = TranslateDTPKRate(curUpdatedSample.rate)
							dtpkOutBytes[curPlayback.GetFileAbsolutePlaybackLocation() + 10:curPlayback.GetFileAbsolutePlaybackLocation() + 11] = array('B', nUpdatedRate.to_bytes(1))
							print ('\t\tUpdated sample table at 0x%x to DTPK rate 0x%x (frequency %u) from 0x%x' % (curPlayback.GetFileAbsolutePlaybackLocation(), nUpdatedRate, curUpdatedSample.rate, thisOldRate))
							if ((previousRate != INVALID_SAMPLE_ID) and (previousRate != thisOldRate) and (not warnedOnce)):
								PrintAsWarning('Warning: this sample is used at multiple rates.  You may want to further adjust the rate value(s) to account for that.')
								warnedOnce = True
							previousRate = thisOldRate

					fMustUpdateHeader = False

				if (fMustUpdateHeader):
					dtpkOutBytes[injectionOffset:injectionOffset + curSampleEntry.GetSampleLength()] = dtpkbytes[curSampleEntry.GetSampleOffset():curSampleEntry.GetSampleOffset() + curSampleEntry.GetSampleLength()]
					if (curSampleEntry.GetSampleOffset() != injectionOffset):
						print ('\tAdjusting sample definition entry at 0x%x for sample location 0x%x length 0x%x (sample was at 0x%x)' % (currentSampleStartingOffset, injectionOffset, curSampleEntry.GetSampleLength(), curSampleEntry.GetSampleOffset()))
						dtpkOutBytes[currentSampleStartingOffset:currentSampleStartingOffset + 4] = array('B', curSampleEntry.GetFlagValueForNewLocation(injectionOffset).to_bytes(4, byteorder='little'))
					injectionOffset += curSampleEntry.GetSampleLength()

			# reset for next pass
			currentSampleStartingOffset += 0x10

		if ((mvc2_DTPK_size_limit < injectionOffset) and fIsProbablyMvC2CharacterFile):
			PrintAsFailure('ERROR: mvc2 character file size limit reached.')

	if ((not fIsShowingAnyInfo) and (len(outFile) == 0) and (not dump_adpcm_samples) and (not dump_pcm_samples_to_wav)):
		print ('\n"%s" is a valid DTPK file.  It contains %u command groups, a total of %u tracks, and uses %u samples (%u PCM, %u ADPCM).' % (dtpkfile, chunkSequencer.GetGroupCount(), chunkSequencer.GetTrackCount(), count_samples, countPCMSamples, countADPCMSamples))
		print ('\tThis script can display detailed file information and/or extract out the samples.')
		print ('\tUse -options to see the current list of options.')
		print ('')

	if (fIsShowingAnyInfo):
		print ('\n')

	if (show_detailed_file_info):
		print ('%u ADPCM samples found' % countADPCMSamples)
	if (countADPCMExports != 0):
		print ('Extracted out %u ADPCM samples' % countADPCMExports)
	elif (dump_adpcm_samples):
		print ('No ADPCM samples found.')

	if (show_detailed_file_info):
		print ('%u PCM samples found' % (countPCMSamples))
	if (countWAVExports != 0):
		print ('Extracted out %u PCM samples to WAV' % countWAVExports)
	elif (dump_pcm_samples_to_wav):
		print ('No PCM samples found.')

	if (len(outFile) != 0):
		# Update the filesize just in case
		if (injectionOffset != filesize_actual):
			# Enforce alignment
			remnant = injectionOffset % 4
			if (remnant != 0):
				print ('\tAdding %u bytes padding (size was 0x%x)' % (remnant, injectionOffset))
				dtpkOutBytes[injectionOffset:injectionOffset + remnant] = array('B', bytes(remnant))
				injectionOffset += remnant
			print ('\tUpdating header: new file size 0x%x' % injectionOffset)
			dtpkOutBytes[8:12] = array('B', injectionOffset.to_bytes(4, byteorder='little'))
			# truncate off the unused region, if any
			while (injectionOffset < len(dtpkOutBytes)):
				dtpkOutBytes.pop()

		if ((injectionOffset > mvc2_DTPK_size_limit) and fIsProbablyMvC2CharacterFile):
			PrintAsFailure('ERROR!  You have 2MB total for audio for Dreamcast.')
			print ('ERROR!  Keep file sizes for mvc2 characters to ~0x%x (%u) bytes to avoid overflow.' % (mvc2_DTPK_size_limit, mvc2_DTPK_size_limit))
			print ('ERROR!  If you cross that boundary, weird things will happen with sounds and/or you\'ll crash the Dreamcast.')
			print ('  If you\'re not worried about MvC2, you\'re probably fine.')
		elif (injectionOffset > (filesize_actual + 20000)):
			PrintAsWarning('Warning: These changes have grown the size of this audio file by %u bytes.  The Dreamcast has extremely limited memory.' % (injectionOffset - filesize_actual))
			PrintAsWarning('As such, downsampling injected files is usually smart if you want to avoid issues.')
		open(outFile, "wb").write(dtpkOutBytes)
		print ('\nFile \'%s\' created.' % outFile)

# =====================================================================================================

def PrintUsage():
	print ('\nAM2 DTPK info script v0.010')
	print ('Usage: python DTPKDump.py [options] <DTPK sound file>')
	print ('Options:')
	print ('          -showinfo         display detailed information about all DTPK contents')
	#print ('          -showheaderinfo   display detailed information about just the header')
	#print ('          -showtrackinfo    display detailed information about just the track data')
	#print ('          -showplaybackinfo display detailed information about just the sample playback data')
	#print ('          -showsampleinfo   display detailed information about just the samples')
	print ('          -extractall       extract out all samples')
	print ('          -wavconv          extract out all samples as PCM WAV files')
	#print ('          -dumpadpcm        extract out all ADPCM samples to raw Yamaha ADPCM files (usable in Awave Studio)')
	#print ('          -dumppcm          extract out all PCM samples to PCM WAV files (may not work for strange playback rates)')
	#print ('          -dumppcmall       extract out *all* samples to PCM WAV files (may not work for strange playback rates)')
	print ('                            (This options uses Sappharad\'s ADPCM to PCM conversion code.)')
	print ('          -spliceadpcm (NUM) (FREQUENCY) (FILENAME)  replace sample NUM in the DTPK with the specified ADPCM file')
	print ('                            Requires -out')
	print ('          -splicepcm (NUM) (FILENAME)  replace sample NUM in the DTPK with the specified PCM WAV file')
	print ('                            WAV files must be PCM 8bit or 16bit at either 11025 or 22050 sample rate')
	print ('                            Requires -out')
	print ('          -out (FILENAME)   save the updated DTPK file as FILENAME')
	print ('                            Requires one of the splicing options.')
	
# =====================================================================================================
if __name__ == '__main__':
	argv = sys.argv
	argc = len(argv)
	os.system('color')

	fFailedArgProcessing = False
	dtpkFilename = ""
	outFile = ""

	changedSamples = []

	if (argc > 1):
		iCurrArg = 1
		while iCurrArg <= len(argv[1:]):
			if (argv[iCurrArg][0] == '-'):
				if (argv[iCurrArg] == "-showinfo"):
					show_detailed_file_info = True
					show_header_info = True
					show_detailed_data_tracks = True
					show_detailed_data_playback = True
					show_detailed_data_samples = True
					show_detailed_data_samples_full = True
				elif (argv[iCurrArg] == "-showheaderinfo"):
					show_header_info = True
				elif (argv[iCurrArg] == "-showtrackinfo"):
					show_detailed_data_tracks = True
				elif (argv[iCurrArg] == "-showplaybackinfo"):
					show_detailed_data_playback = True
				elif (argv[iCurrArg] == "-showeffectsinfo"):
					show_detailed_data_effects = True
				elif (argv[iCurrArg] == "-showsampleinfo"):
					show_detailed_data_samples = True
					show_detailed_data_samples_full = True
				elif (argv[iCurrArg] == "-reassemble"):
					reassembly_test = True
				elif ((argv[iCurrArg] == "-dump") or (argv[iCurrArg] == "-dumpall") or (argv[iCurrArg] == "-extractall")):
					dump_adpcm_samples = True
					dump_pcm_samples_to_wav = True
				elif (argv[iCurrArg] == "-dumpadpcm"):
					dump_adpcm_samples = True
				elif (argv[iCurrArg] == "-dumppcm"):
					dump_pcm_samples_to_wav = True
				elif ((argv[iCurrArg] == "-dumppcmall") or (argv[iCurrArg] == "-wavconv")):
					dump_pcm_samples_to_wav = True
					dump_adpcm_samples_to_wav = True
				elif (argv[iCurrArg] == "-spliceadpcm"):
					iCurrArg += 1
					try:
						samplePosition = int(argv[iCurrArg])
					except:
						try:
							samplePosition = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						sampleRate = int(argv[iCurrArg])
					except:
						try:
							sampleRate = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						adpcmFile = argv[iCurrArg]
					except:
						PrintAsFailure('ERROR: check usage.  You need to tell us what ADPCM file to splice in.')
						fFailedArgProcessing = True
						break
					
					if (not ImportAndAddADPCMFile(adpcmFile, changedSamples, samplePosition, sampleRate)):
						PrintAsFailure('ERROR: file \'%s\' could not be imported.' % adpcmFile)
						fFailedArgProcessing = True
						break
				elif (argv[iCurrArg] == "-splicepcm"):
					iCurrArg += 1
					try:
						samplePosition = int(argv[iCurrArg])
					except:
						try:
							samplePosition = int(argv[iCurrArg], 16)
						except:
							PrintAsFailure('ERROR: check usage.  You need to tell us what sample slot to inject this sample into.\n')
							fFailedArgProcessing = True
							break
					iCurrArg += 1
					try:
						wavFile = argv[iCurrArg]
					except:
						PrintAsFailure('ERROR: check usage.  You need to tell us what WAV file to splice in.')
						fFailedArgProcessing = True
						break
					
					if (not ImportAndAddWAVFile(wavFile, changedSamples, samplePosition)):
						PrintAsFailure('ERROR: file \'%s\' could not be imported.' % wavFile)
						fFailedArgProcessing = True
						break
				elif ((argv[iCurrArg] == "-o") or (argv[iCurrArg] == "-out")):
					iCurrArg += 1
					outFile = argv[iCurrArg]
				elif ((argv[iCurrArg] == "-options") or (argv[iCurrArg] == "-?")):
					break
				else:
					PrintAsFailure('Error: unsupported option %s' % argv[iCurrArg])
					fFailedArgProcessing = True
					break
			else:
				dtpkFilename = argv[iCurrArg]
			iCurrArg += 1

	if ((len(changedSamples) != 0) and (len(outFile) == 0)):
		PrintAsFailure('ERROR: you must specify an output filename if trying to splice in a sample.\n')
		dtpkFilename = ""
	if ((len(changedSamples) == 0) and (len(outFile) != 0)):
		PrintAsFailure('ERROR: output file specified with no samples to splice in.\n')
		dtpkFilename = ""

	if ((len(dtpkFilename) != 0) and (not fFailedArgProcessing)):
		dtpkparse(dtpkFilename, outFile, changedSamples)
	else:
		PrintUsage()
		sys.exit(0)

# =====================================================================================================

# =====================================================================================================
# 24-05-31 (0.010)   - Stateful tracking of file data
# 24-05-29 (0.009.x) - Better support for VTennis, DOA2, SPD oddities, and know about the VTennis mystery table
# 24-05-18 (0.009)   - Mapped out more of the DTPK structure
# 24-04-25 (0.008)   - Improve handling of sequencer groups
# 24-04-20 (0.007)   - Allow exporting YADPCM chunks as RIFF/WAV PCM
# 24-04-16 (0.006)   - Significantly improve rate handling
# 24-04-13 (0.005)   - Efficiency: always inline sample replacement
# 24-04-10 (0.004)   - Support splicing in ADPCM chunks or PCM WAV files
# 24-04-03 (0.003)   - Support slightly limited PCM export to WAV
# 24-03-27 (0.002)   - Smooth out T->SPD->S logic, filenames, etc.
# 24-03-21 (0.001)    - Add optional sample extraction
# 24-03-01 (0.000)   - Initial version based upon KingShriek and Shenmue's knowledge.
# =====================================================================================================
